from __future__ import print_function

__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import sys # Provides access to some variables used or maintained by the interpreter and to functions that interact strongly with the interpreter. It is always available.
from cerebrum.crossmodal import MapperUtil # BUILT-IN Crosmodal operations package
from cerebrum.hearing import HearingPerception, HearingMemoryUtil # BUILT-IN Hearing Memory perception package
from cerebrum.vision import VisionPerception, VisionMemoryUtil # BUILT-IN Vision Memory operations package
import time # Provides various time-related functions.
import pyaudio
import cv2 # (Open Source Computer Vision) is a library of programming functions mainly aimed at real-time computer vision.
import numpy # The fundamental package for scientific computing with Python.
from cerebrum.neuralnet.datasets import SequentialDataSet
from itertools import cycle
from pybrain.tools.shortcuts import buildNetwork
from pybrain.structure.modules import LSTMLayer
from pybrain.supervised import RPropMinusTrainer
from sys import stdout
import matplotlib.pyplot as plt

CHUNK = 1024 # Smallest unit of audio. 1024 bytes
FORMAT = pyaudio.paInt16 # Data format
CHANNELS = 2 # Number of channels
RATE = 44100 # Bit Rate of audio stream / Frame Rate

class NeuralWeaver():

	# MAIN CODE BLOCK
	@staticmethod
	def start():
		pairs = MapperUtil.get_allpairs() # Get pairs starting from 0th line
		if not pairs:
			print ("No pairs found.")
			sys.exit()

		p = pyaudio.PyAudio() # Create a PyAudio session

		# Create a stream
		stream = p.open(format=FORMAT,
						channels=CHANNELS,
						rate=RATE,
						output=True)

		h2v_net = buildNetwork(2048, 5, 230400, hiddenclass=LSTMLayer, outputbias=False, recurrent=True)
		data = [1] * 3 + [2] * 3
		# Loop over the pairs coming from CROSSMODAL
		for pair in pairs:
			   #time.sleep(0.5) # Wait 0.5 seconds to prevent aggressive loop
			   if pair['direction'] == "H2V":
					ds = SequentialDataSet(2048, 230400)
					hearing_memory = HearingMemoryUtil.get_memory(pair['timestamp1'])
					vision_memory = VisionMemoryUtil.get_memory(pair['timestamp2'])
					if not vision_memory:
						continue
					for chunky in hearing_memory['data']:
						chunky_array = numpy.fromstring(chunky, 'int16')
						print (chunky_array)
						ds.addSample(chunky_array, numpy.fromstring(vision_memory['amodal'][0], numpy.uint8))
						stream.write(chunky)
					#print len(vision_memory['amodal'])
					for frame in vision_memory['amodal']:
						frame = numpy.fromstring(frame, numpy.uint8).reshape(360,640)
						#print frame.shape
						cv2.imshow("Frame Threshhold", frame)
						cv2.moveWindow("Frame Threshhold",50,100)
						key = cv2.waitKey(1) & 0xFF
					for frame in vision_memory['color']:
						frame = numpy.fromstring(frame, numpy.uint8).reshape(360,640,3)
						#print frame.shape
						cv2.imshow("Frame Delta Colored", frame)
						cv2.moveWindow("Frame Delta Colored",1200,100)
						key = cv2.waitKey(1) & 0xFF

					trainer = RPropMinusTrainer(h2v_net, dataset=ds)
					train_errors = [] # save errors for plotting later
					EPOCHS_PER_CYCLE = 5
					CYCLES = 100
					EPOCHS = EPOCHS_PER_CYCLE * CYCLES
					for i in xrange(CYCLES):
						t0 = time.time() # Initiation time
						trainer.trainEpochs(EPOCHS_PER_CYCLE)
						train_errors.append(trainer.testOnData())
						epoch = (i+1) * EPOCHS_PER_CYCLE
						#print (time.time() - t0) # Elapsed time
						print("\r epoch {}/{} --- {}".format(epoch, EPOCHS, (time.time() - t0)), end="")
						stdout.flush()

					print()
					print("final error =", train_errors[-1])



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import rethinkdb as r # Rethinkdb Python driver

# Memory class
class Memory(object):
	def __init__(self, starting_time, ending_time, data): # Initialize the object
		self.starting_time = starting_time # Starting time attribute
		self.ending_time = ending_time # Ending time attribute
		self.data = data # Data attribute

# Timestamp class
class Timestamp(object):
	def __init__(self, starting_time, ending_time): # Initialize the object
		self.starting_time = starting_time # Starting time attribute
		self.ending_time = ending_time # Ending time attribute

# Convert object to dictionary
def makeit_dict(obj):
	if isinstance(obj, set):
		return list(obj)
	return obj.__dict__

class LanguageMemoryUtil():

	# Add a memory function
	@staticmethod
	def add_memory(data, starting_time, ending_time):
		conn = r.connect("localhost", 28015)
		r.db('test').table("language_memory").insert([
			{ "starting_time": starting_time.strftime("%Y-%m-%d %H:%M:%S.%f"),
			  "ending_time": ending_time.strftime("%Y-%m-%d %H:%M:%S.%f"),
			  "data": r.binary(data)
			}
		]).run(conn)
		r.db('test').table("language_timestamps").insert([
			{ "starting_time": starting_time.strftime("%Y-%m-%d %H:%M:%S.%f"),
			  "ending_time": ending_time.strftime("%Y-%m-%d %H:%M:%S.%f")
			}
		]).run(conn)
		conn.close()

	# Get a memory function
	@staticmethod
	def get_memory(starting_time):
		conn = r.connect("localhost", 28015)
		cursor = r.db('test').table("language_memory").filter({'starting_time': starting_time}).run(conn)
		conn.close()
		return cursor

	# Get timestamps function
	@staticmethod
	def get_timestamps():
		conn = r.connect("localhost", 28015)
		cursor = r.db('test').table("language_timestamps").run(conn)
		conn.close()
		return cursor



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import pyaudio # Provides Python bindings for PortAudio, the cross platform audio API
import wave # Provides a convenient interface to the WAV sound format
import datetime # Supplies classes for manipulating dates and times in both simple and complex ways
import os.path # The path module suitable for the operating system Python is running on, and therefore usable for local paths
import audioop # Operates on sound fragments consisting of signed integer samples 8, 16 or 32 bits wide, stored in Python strings.
import numpy # The fundamental package for scientific computing with Python.
import multiprocessing # A package that supports spawning processes using an API similar to the threading module.
import imutils # A series of convenience functions to make basic image processing functions such as translation, rotation, resizing, skeletonization etc.
import pyqtgraph as pg # A pure-python graphics and GUI library built on PyQt4 / PySide and numpy
from PyQt4 import QtCore, QtGui # A comprehensive set of Python bindings for Digia's Qt cross platform GUI toolkit.
import time # Provides various time-related functions.
from cerebrum.hearing.utilities import HearingMemoryUtil # BUILT-IN Memory operations package
import Tkinter

CHUNK = 1024 # Smallest unit of audio. 1024 bytes
FORMAT = pyaudio.paInt16 # Data format
CHANNELS = 2 # Number of channels
RATE = 44100 # Bit Rate of audio stream / Frame Rate
THRESHOLD = 1000 # Threshhold value for detecting stimulant
SILENCE_DETECTION = 5 # Wait number of frames to decide whether it fell silent or not
EMPTY_CHUNK = chr(int('000000', 2)) * CHUNK * 4 # Create an empty unit of audio for just once
WAVE_OUTPUT_FILENAME = "hearing/memory/" +  str(datetime.date.today()) + ".wav" # Example path if saving needed
root = Tkinter.Tk()
SCREEN_WIDTH = root.winfo_screenwidth()
SCREEN_HEIGHT = root.winfo_screenheight()

class HearingPerception():

	# A function that will save recordings to a file
	@staticmethod
	def save_file():
		if not os.path.isfile(WAVE_OUTPUT_FILENAME): # If there is not such a file
			wf = wave.open(WAVE_OUTPUT_FILENAME, 'wb') # Create the file
			wf.setnchannels(CHANNELS) # Set number of channels
			wf.setsampwidth(p.get_sample_size(FORMAT)) # Set sampling format
			wf.setframerate(RATE) # Set Bit Rate / Frame Rate
			wf.writeframes("") # Write nothing
			wf.close() # Close the session

		wf = wave.open(WAVE_OUTPUT_FILENAME, 'rb') # Open the file with only read permission
		n_frames = wf.getnframes() # Get all frames in it
		previous_wav = wf.readframes(n_frames) # Assign all frames to a variable
		wf.close() # Close the session

		wf = wave.open(WAVE_OUTPUT_FILENAME, 'wb') # Open the file with write permission
		wf.setnchannels(CHANNELS) # Set number of channels
		wf.setsampwidth(p.get_sample_size(FORMAT)) # Set sampling format
		wf.setframerate(RATE) # Set Bit Rate / Frame Rate
		wf.writeframes(previous_wav + b''.join(frames)) # Write the all frames including previous ones
		wf.close() # Close the session

	# A function that will compute frequency of chunk with Fourier Transform
	@staticmethod
	def find_frequency(data):
		T = 1.0/RATE # Reciprocal of Bit Rate
		N = data.shape[0] # Number of rows in data(numpy array)
		Pxx = (1./N)*numpy.fft.fft(data) # Compute the one-dimensional n-point discrete Fourier Transform (DFT) of data with the efficient Fast Fourier Transform (FFT) algorithm [CT]
		f = numpy.fft.fftfreq(N,T) # Return the Discrete Fourier Transform sample frequencies
		Pxx = numpy.fft.fftshift(Pxx) # Shift the zero-frequency component to the center of the spectrum
		f = numpy.fft.fftshift(f) # Shift the zero-frequency component to the center of the spectrum
		return f, Pxx # Return the results

	# A function that will draw a spectrum analyzer graphic to screen (PyQtGraph)
	@staticmethod
	def draw_spectrum_analyzer(all_frames, thresh_frames):
		time.sleep(1) # Wait just one second
		pw = pg.plot(title="Spectrum Analyzer") # Window title
		pg.setConfigOptions(antialias=True) # Enable antialias for better resolution
		pw.win.resize(800, 300) # Define window size
		pw.win.move(540 * SCREEN_WIDTH / 1920, 500 * SCREEN_HEIGHT / 1080) # Define window position
		while True: # Loop over the frames of the audio / data chunks
			data = ''.join(all_frames[-1:]) # Get only the last frame of all frames
			data = numpy.fromstring(data, 'int16') # Binary string to numpy int16 data format
			pw.setMouseEnabled(y=False) # Disable mouse
			pw.setYRange(0,1000) # Set Y range of graph
			pw.setXRange(-(RATE/16), (RATE/16), padding=0) # Set X range of graph relative to Bit Rate
			pwAxis = pw.getAxis("bottom") # Get bottom axis
			pwAxis.setLabel("Frequency [Hz]") # Set bottom axis label
			f, Pxx = HearingPerception.find_frequency(data) # Call find frequency function
			f = f.tolist() # Numpy array to list
			Pxx = (numpy.absolute(Pxx)).tolist() # Numpy array to list
			try: # Try this block
				if thresh_frames[-1:][0] == EMPTY_CHUNK: # If last thresh frame is equal to EMPTY CHUNK
					pw.plot(x=f,y=Pxx, clear=True, pen=pg.mkPen('w', width=1.0, style=QtCore.Qt.SolidLine)) # Then plot with white pen
				else: # If last thresh frame is not equal to EMPTY CHUNK
					pw.plot(x=f,y=Pxx, clear=True, pen=pg.mkPen('y', width=1.0, style=QtCore.Qt.SolidLine)) # Then plot with yellow pen
			except IndexError: # If we are getting an IndexError because of this -> thresh_frames[-1:][0]
				pw.plot(x=f,y=Pxx, clear=True, pen=pg.mkPen('w', width=1.0, style=QtCore.Qt.SolidLine)) # Then plot with white pen
			pg.QtGui.QApplication.processEvents() # ???
			time.sleep(0.05) # Wait a few miliseconds

	# A function that will draw a waveform graphic to screen (PyQtGraph)
	@staticmethod
	def draw_waveform(all_frames, thresh_frames):
		time.sleep(1) # Wait just one second
		pw = pg.plot(title="Waveform") # Window title
		pg.setConfigOptions(antialias=True) # Enable antialias for better resolution
		pw.win.resize(1300, 160) # Define window size
		pw.win.move(300 * SCREEN_WIDTH / 1920, 850 * SCREEN_HEIGHT / 1080) # Define window position
		pw.showAxis('bottom', False) # Hide bottom axis
		while True: # Loop over the frames of the audio / data chunks
			data = ''.join(all_frames[-20:]) # Join last 20 frames of all frames
			data = numpy.fromstring(data, 'int16') # Binary string to numpy int16 data format
			data2 = ''.join(thresh_frames[-20:]) # Join last 20 frames of thrsh frames
			data2 = numpy.fromstring(data2, 'int16') # Binary string to numpy int16 data format
			pw.setMouseEnabled(x=False) # Disable mouse
			pw.setRange(yRange=[-10000,10000]) # Set Y range of graph
			pw.plot(data, clear=True, pen=pg.mkPen('w', width=0.5, style=QtCore.Qt.DotLine)) # Plot all frames with white pen
			pw.plot(data2, pen=pg.mkPen('y', width=0.5, style=QtCore.Qt.DotLine)) # Plot thresh frames with yellow pen
			text = pg.TextItem("Seconds : " + str(int(len(all_frames)/(RATE/CHUNK))), color=(255, 255, 255)) # Define seconds according to number of total frames as a text
			pw.addItem(text) # Display seconds according to number of total frames
			text.setPos(500, 0) # Set text position
			pg.QtGui.QApplication.processEvents()
			time.sleep(0.05) # Wait a few miliseconds

	# MAIN CODE BLOCK
	@staticmethod
	def start(audio_input, hearing_perception_stimulated):

		if audio_input == "0":
			pass
		else:
			wf = wave.open(audio_input, 'rb') # Open .wav file from given path as audio_input in arguments

		p = pyaudio.PyAudio() # Create a PyAudio session

		# Create a stream
		if audio_input == "0":
			stream = p.open(format=FORMAT,
						channels=CHANNELS,
						rate=RATE,
						input=True,
						frames_per_buffer=CHUNK)
		else:
			stream = p.open(format=p.get_format_from_width(wf.getsampwidth()),
						channels=wf.getnchannels(),
						rate=wf.getframerate(),
						output=True)

		hearing_manager = multiprocessing.Manager() # Shared memory space manager
		memory_data = [] # Define memory data array
		all_frames = hearing_manager.list() # Define all_frames array in shared memory
		thresh_frames = hearing_manager.list() # Define thresh_frames array in shared memory

		if audio_input == "0":
			data = stream.read(CHUNK) # Get first data frame from .wav file
		else:
			data = wf.readframes(CHUNK) # Get first data frame from .wav file

		all_frames.append(data) # Append to all frames
		thresh_frames.append(EMPTY_CHUNK) # Append an EMPTY CHUNK to thresh frames

		process1 = multiprocessing.Process(target=HearingPerception.draw_waveform, args=(all_frames, thresh_frames)) # Define draw waveform process
		process1.start() # Start draw waveform process

		process2 = multiprocessing.Process(target=HearingPerception.draw_spectrum_analyzer, args=(all_frames, thresh_frames)) # Define draw spectrum analyzer process
		process2.start() # Start drar spectrum analyzer process

		# Loop over the frames of the audio / data chunks
		while data != '':
			previous_data = data # Get previous chunk that coming from end of the loop

			if audio_input == "0":
				data = stream.read(CHUNK) # Read a new chunk from the stream
			else:
				stream.write(data) # Monitor current chunk
				data = wf.readframes(CHUNK) # Read a new chunk from the stream

			all_frames.append(data) # Append this chunk to all frames
			thresh_frames.append(EMPTY_CHUNK) # Append an EMPTY CHUNK to thresh frames

			rms = audioop.rms(data, 2) # Calculate Root Mean Square of current chunk
			if rms >= THRESHOLD: # If Root Mean Square value is greater than THRESHOLD constant
				starting_time = datetime.datetime.now() # Starting time of the memory
				hearing_perception_stimulated.value = 1 # Hearing perception stimulated
				thresh_frames.pop() # Pop out last frame of thresh frames
				thresh_frames.pop() # Pop out last frame of thresh frames
				memory_data.append(previous_data) # Append previous chunk to memory data
				thresh_frames.append(previous_data) # APpend previos chunk to thresh frames
				memory_data.append(data) # Append current chunk to memory data
				thresh_frames.append(data) # Append current chunk to thresh frames
				silence_counter = 0 # Define silence counter
				while silence_counter < SILENCE_DETECTION: # While silence counter value less than SILENCE_DETECTION constant

					if audio_input == "0":
						data = stream.read(CHUNK) # Read a new chunk from the stream
					else:
						stream.write(data) # Monitor current chunk
						data = wf.readframes(CHUNK) # Read a new chunk from the stream

					all_frames.append(data) # Append this chunk to all frames
					memory_data.append(data) # Append this chunk to memory data
					thresh_frames.append(data) # Append this chunk to thresh frames
					rms = audioop.rms(data, 2) # Calculate Root Mean Square of current chunk again

					if rms < THRESHOLD: # If Root Mean Square value is less than THRESHOLD constant
						silence_counter += 1 # Then increase silence counter
					else: # Else
						silence_counter = 0 # Assign zero value to silence counter

				del memory_data[-(SILENCE_DETECTION-2):] # Delete last frames of memory data as much as SILENCE_DETECTION constant
				del thresh_frames[-(SILENCE_DETECTION-2):] # Delete last frames of thresh frames as much as SILENCE_DETECTION constant
				for i in range(SILENCE_DETECTION-2): # SILENCE_DETECTION constant times
					thresh_frames.append(EMPTY_CHUNK) # Append an EMPTY_CHUNK
				ending_time = datetime.datetime.now() # Ending time of the memory
				hearing_perception_stimulated.value = 0 # Hearing perception NOT stimulated

				memory_data = ''.join(memory_data)
				#HearingMemoryUtil.add_memory(memory_data, starting_time, ending_time)
				process3 = multiprocessing.Process(target=HearingMemoryUtil.add_memory, args=(memory_data, starting_time, ending_time)) # Define write memory process
				process3.start() # Start write memory process
				memory_data = [] # Empty memory data

		process1.terminate() # Terminate draw waveform process
		process2.terminate() # Terminate drar spectrum analyzer process
		stream.stop_stream() # Stop the stream
		stream.close() # Close the stream
		p.terminate() # Terminate the session



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import rethinkdb as r # Rethinkdb Python driver

# Pair class
class Pair(object):
	def __init__(self, timestamp1, timestamp2, direction): # Initialize the object
		self.timestamp1 = timestamp1 # Memory starting time of a sense
		self.timestamp2 = timestamp2 # Memory starting time of another sense
		self.direction = direction # Direction. For example H2V, V2H, H2H etc.

# Convert object to dictionary
def makeit_dict(obj):
	if isinstance(obj, set):
		return list(obj)
	return obj.__dict__

class MapperUtil():

	# Add a pair function
	@staticmethod
	def add_pair(timestamp1, timestamp2, direction):
		conn = r.connect("localhost", 28015)
		r.db('test').table("crossmodal_mappings").insert([
			{ "timestamp1": timestamp1,
			  "timestamp2": timestamp2,
			  "direction": direction
			}
		]).run(conn)
		conn.close()

	# Get a pair function
	@staticmethod
	def get_pair_by_direction(direction):
		conn = r.connect("localhost", 28015)
		cursor = r.db('test').table("crossmodal_mappings").filter({'direction': direction}).run(conn)
		conn.close()
		return cursor

	# Get all pairs function
	@staticmethod
	def get_allpairs():
		conn = r.connect("localhost", 28015)
		cursor = r.db('test').table("crossmodal_mappings").run(conn)
		conn.close()
		return cursor



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import argparse # Makes it easy to write user-friendly command-line interfaces.
import multiprocessing # A package that supports spawning processes using an API similar to the threading module.
from cerebrum.hearing import HearingPerception # Hearing Package
from cerebrum.vision import VisionPerception # Vision Package
from cerebrum.language import LanguageAnalyzer # Language Package
from cerebrum.crossmodal import MapperStarters # Crossmodal Package
#from cerebrum.neuralnet import NeuralWeaver # NeuralNet Package
import time
from distutils.dir_util import mkpath
import os.path
import os
import subprocess
import rethinkdb as r # Rethinkdb Python driver

def initiate():
	ap = argparse.ArgumentParser() # Define an Argument Parser
	ap.add_argument("-v", "--video", help="path to the video file") # Add --video argument
	ap.add_argument("-a", "--audio", help="path to the audio file") # Add --audio argument
	ap.add_argument("-c", "--captions", help="path to the captios file") # Add --captions argument
	args = vars(ap.parse_args()) # Parse the arguments

	stem_manager = multiprocessing.Manager() # Shared memory space manager
	hearing_perception_stimulated = stem_manager.Value('i', 0) # Define hearing perception stimualted variable in shared memory to get if it's stimulated or not (Integer)
	vision_perception_stimulated = stem_manager.Value('i', 0) # Define vision perception stimualted variable in shared memory to get if it's stimulated or not (Integer)
	language_analysis_stimulated = stem_manager.Value('i', 0) # Define language analysis stimualted variable in shared memory to get if it's stimulated or not (Integer)

	active_perceptions = 0

	#args = [os.path.expanduser("--directory ~/ComeOnRethink")]
	#os.execvp("rethinkdb", args)
	subprocess.Popen(['rethinkdb', '--directory', os.path.expanduser('~/Hippocampus')]) # RethinkDB directory to store data and metadata
	time.sleep(3)
	conn = r.connect("localhost", 28015)
	try:
		r.db('test').table_create('hearing_memory').run(conn)
	except:
		pass
	try:
		r.db('test').table_create('hearing_timestamps').run(conn)
	except:
		pass
	try:
		r.db('test').table_create('language_memory').run(conn)
	except:
		pass
	try:
		r.db('test').table_create('language_timestamps').run(conn)
	except:
		pass
	try:
		r.db('test').table_create('vision_memory').run(conn)
	except:
		pass
	try:
		r.db('test').table_create('vision_timestamps').run(conn)
	except:
		pass
	try:
		r.db('test').table_create('crossmodal_mappings').run(conn)
	except:
		pass
	conn.close()
	time.sleep(3)

	if args["audio"] is None:
		pass
	else:
		hearing_perception_process = multiprocessing.Process(target=HearingPerception.start, args=(args["audio"],hearing_perception_stimulated)) # Define hearing perception process
		hearing_perception_process.start() # Start hearing perception process
		active_perceptions += 1

	if args["video"] is None:
		pass
	else:
		vision_perception_process = multiprocessing.Process(target=VisionPerception.start, args=(args["video"],vision_perception_stimulated)) # Define vision perception process
		vision_perception_process.start() # Start vision perception process
		active_perceptions += 1

	if args["captions"] is None:
		pass
	else:
		language_analysis_process = multiprocessing.Process(target=LanguageAnalyzer.start, args=(args["captions"],language_analysis_stimulated)) # Define language analysis process
		language_analysis_process.start() # Start language analysis process
		active_perceptions += 1

	crossmodal_mapperHV_process = multiprocessing.Process(target=MapperStarters.startHV) # Define crossmodal mapper for hearing & vision process
	crossmodal_mapperHV_process.start() # Start crossmodal mapperHV process

	crossmodal_mapperHL_process = multiprocessing.Process(target=MapperStarters.startHL) # Define crossmodal mapper for hearing & language process
	crossmodal_mapperHL_process.start() # Start crossmodal mapperHL process

	crossmodal_mapperVL_process = multiprocessing.Process(target=MapperStarters.startVL) # Define crossmodal mapper for vision & language process
	crossmodal_mapperVL_process.start() # Start crossmodal mapperVL process

	training = 0

	while True:
		if args["audio"]:
			if not hearing_perception_process.is_alive():
				active_perceptions -= 1
				args["audio"] = None
				print "WARNING: Hearing Perception process is terminated."
		if args["video"]:
			if not vision_perception_process.is_alive():
				active_perceptions -= 1
				args["video"] = None
				print "WARNING: Vision Perception process is terminated."
		if args["captions"]:
			if not language_analysis_process.is_alive():
				active_perceptions -= 1
				args["captions"] = None
				print "WARNING: Language Analysis process is terminated."
		if active_perceptions == 0 and not training:
				#neuralnet_weaver_process = multiprocessing.Process(target=NeuralWeaver.start) # Define neuralnet weaver process
				#neuralnet_weaver_process.start() # Start neuralnet weaver process
				training = 1
		if training and not neuralnet_weaver_process.is_alive():
			crossmodal_mapper_process.terminate()
			print "EXIT: Training is finished."
			break



'''An implementation of several types of Restricted Boltzmann Machines.

This code is largely based on the Matlab generously provided by Taylor, Hinton
and Roweis, and described in their 2006 NIPS paper, "Modeling Human Motion Using
Binary Hidden Variables". Their code and results are available online at
http://www.cs.nyu.edu/~gwtaylor/publications/nips2006mhmublv/.

There are more RBM implementations in this module than just the Taylor
"conditional RBM," though. The basic (non-Temporal) RBM is based on the Taylor,
Hinton, and Roweis code, but stripped of the dynamic bias terms and refactored
into an object-oriented style. The convolutional RBM code is based on the 2009
ICML paper by Lee, Grosse, Ranganath and Ng, "Convolutional Deep Belief Networks
for Scalable Unsupervised Learning of Hierarchical Representations".

The mean-covariance RBM is based on XXX.

All implementations incorporate an option to train hidden unit biases using a
sparsity criterion, as described in the 2008 NIPS paper by Lee, Ekanadham and
Ng, "Sparse Deep Belief Net Model for Visual Area V2".

Most RBM implementations  provide an option to treat visible units as either
binary or gaussian. Training networks with gaussian visible units is a tricky
dance of parameter-twiddling, but binary units seem quite stable in their
learning and convergence properties.

Finally, although I have tried to ensure that the code is correct, there are
probably many bugs, all of which are my own doing. I wrote this code to get a
better intuitive understanding for the RBM family of machine learning
algorithms, but I do not claim that the code is useful for a particular purpose
or produces state-of-the-art results. Mostly I hope that this code is readable
so that others can use it to better understand how this whole RBM thing works.
'''

import numpy as np
import logging
import numpy.random as rng


def sigmoid(eta):
	'''Return the logistic sigmoid function of the argument.'''
	return 1. / (1. + np.exp(-eta))


def identity(eta):
	'''Return the identity function of the argument.'''
	return eta


def bernoulli(p):
	'''Return an array of boolean samples from Bernoulli(p).

	Parameters
	----------
	p : ndarray
		This array should contain values in [0, 1].

	Returns
	-------
	An array of the same shape as p. Each value in the result will be a boolean
	indicating whether a single Bernoulli trial succeeded for the corresponding
	element of `p`.
	'''
	return rng.rand(*p.shape) < p


class RBM(object):
	'''A Restricted Boltzmann Machine (RBM) is a probabilistic model of data.

	RBMs have two layers of variables (here called "units," in keeping with
	neural network terminology) -- a "visible" layer that models data in the
	world, and a "hidden" layer that is imagined to generate the data in the
	visible layer. RBMs are inspired by the (unrestricted) Boltzmann Machine, a
	model from statistical physics in which each unit is connected to all other
	units, and the states of unobserved variables can be inferred by using a
	sampling procedure.

	The full connectivity of an unrestricted Boltzmann Machine makes inference
	difficult, requiring sampling or some other approximate technique. RBMs
	restrict this more general model by requiring that the visible and hidden
	units form a fully connected, undirected, bipartite graph. In this way, each
	of the visible units is independent from the other visible units when
	conditioned on the state of the hidden layer, and each of the hidden units
	is independent of the others when conditioned on the state of the visible
	layer. This conditional independence makes inference tractable for the units
	in a single RBM.

	To "encode" a signal by determining the state of the hidden units given some
	visible data ("signal"),

	1. the signal is presented to the visible units, and
	2. the states of the hidden units are sampled from the conditional
	   distribution given the visible data.

	To "decode" an encoding in the hidden units,

	3. the states of the visible units are sampled from the conditional
	   distribution given the states of the hidden units.

	Once a signal has been encoded and then decoded,

	4. the sampled visible units can be compared directly with the original
	   visible data.

	Training takes place by presenting a number of data points to the network,
	encoding the data, reconstructing it from the hidden states, and encoding
	the reconstruction in the hidden units again. Then, using contrastive
	divergence (Hinton 2002; Hinton & Salakhutdinov 2006), the gradient is
	approximated using the correlations between visible and hidden units in the
	first encoding and the same correlations in the second encoding.
	'''

	def __init__(self, num_visible, num_hidden, binary=True, scale=0.001):
		'''Initialize a restricted boltzmann machine.

		Parameters
		----------
		num_visible : int
			The number of visible units.

		num_hidden : int
			The number of hidden units.

		binary : bool
			True if the visible units are binary, False if the visible units are
			normally distributed.

		scale : float
			Sample initial weights from N(0, scale).
		'''
		self.weights = scale * rng.randn(num_hidden, num_visible)
		self.hid_bias = scale * rng.randn(num_hidden, 1)
		self.vis_bias = scale * rng.randn(num_visible, 1)

		self._visible = binary and sigmoid or identity

	@property
	def num_hidden(self):
		return len(self.hid_bias)

	@property
	def num_visible(self):
		return len(self.vis_bias)

	def hidden_expectation(self, visible, bias=0.):
		'''Given visible data, return the expected hidden unit values.'''
		return sigmoid(np.dot(self.weights, visible.T) + self.hid_bias + bias)

	def visible_expectation(self, hidden, bias=0.):
		'''Given hidden states, return the expected visible unit values.'''
		return self._visible(np.dot(hidden.T, self.weights) + self.vis_bias.T + bias)

	def iter_passes(self, visible):
		'''Repeatedly pass the given visible layer up and then back down.

		Parameters
		----------
		visible : ndarray
			The initial state of the visible layer.

		Returns
		-------
		Generates a sequence of (visible, hidden) states. The first pair will be
		the (original visible, resulting hidden) states, followed by pairs
		containing the values from (visible down-pass, hidden up-pass).
		'''
		while True:
			hidden = self.hidden_expectation(visible)
			yield visible, hidden
			visible = self.visible_expectation(bernoulli(hidden))

	def reconstruct(self, visible, passes=1):
		'''Reconstruct a given visible layer through the hidden layer.

		Parameters
		----------
		visible : ndarray
			The initial state of the visible layer.

		passes : int
			The number of up- and down-passes.

		Returns
		-------
		An array containing the reconstructed visible layer after the specified
		number of up- and down- passes.
		'''
		for i, (visible, _) in enumerate(self.iter_passes(visible)):
			if i + 1 == passes:
				return visible


class Trainer(object):
	'''
	'''

	def __init__(self, rbm, momentum=0., l2=0., target_sparsity=None):
		'''
		'''
		self.rbm = rbm
		self.momentum = momentum
		self.l2 = l2
		self.target_sparsity = target_sparsity

		self.grad_weights = np.zeros(rbm.weights.shape, float)
		self.grad_vis = np.zeros(rbm.vis_bias.shape, float)
		self.grad_hid = np.zeros(rbm.hid_bias.shape, float)

	def learn(self, visible, learning_rate=0.2):
		'''
		'''
		gradients = self.calculate_gradients(visible)
		self.apply_gradients(*gradients, learning_rate=learning_rate)

	def calculate_gradients(self, visible_batch):
		'''Calculate gradients for a batch of visible data.

		Returns a 3-tuple of gradients: weights, visible bias, hidden bias.

		visible_batch: A (batch size, visible units) array of visible data. Each
		  row represents one visible data sample.
		'''
		passes = self.rbm.iter_passes(visible_batch)
		v0, h0 = passes.next()
		v1, h1 = passes.next()
		gw = (np.dot(h0, v0) - np.dot(h1, v1)) / len(visible_batch)
		gv = (v0 - v1).mean(axis=0)
		gh = (h0 - h1).mean(axis=1)
		if self.target_sparsity is not None:
			gh = self.target_sparsity - h0.mean(axis=1)

		logging.debug('displacement: %.3g, hidden std: %.3g',
					  np.linalg.norm(gv), h0.std(axis=1).mean())
		# make sure we pass ndarrays
		gv = gv.reshape(gv.shape[0],1)
		gh = gh.reshape(gh.shape[0],1)
		return gw, gv, gh

	def apply_gradients(self, weights, visible, hidden, learning_rate=0.2):
		'''
		'''
		def update(name, g, _g, l2=0):
			target = getattr(self.rbm, name)
			g *= 1 - self.momentum
			g += self.momentum * (g - l2 * target)
			target += learning_rate * g
			_g[:] = g

		update('vis_bias', visible, self.grad_vis)
		update('hid_bias', hidden, self.grad_hid)
		update('weights', weights, self.grad_weights, self.l2)


class Temporal(RBM):
	'''An RBM that incorporates temporal (dynamic) visible and hidden biases.

	This RBM is based on work and code by Taylor, Hinton, and Roweis (2006).
	'''

	def __init__(self, num_visible, num_hidden, order, binary=True, scale=0.001):
		'''
		'''
		super(TemporalRBM, self).__init__(
			num_visible, num_hidden, binary=binary, scale=scale)

		self.order = order

		self.vis_dyn_bias = scale * rng.randn(order, num_visible, num_visible)
		self.hid_dyn_bias = scale * rng.randn(order, num_hidden, num_visible) - 1.

	def iter_passes(self, frames):
		'''Repeatedly pass the given visible layer up and then back down.

		Generates the resulting sequence of (visible, hidden) states.

		visible: An (order, visible units) array containing frames of visible
		  data to "prime" the network. The temporal order of the frames is
		  assumed to be reversed, so frames[0] will be the current visible
		  frame, frames[1] is the previous frame, etc.
		'''
		vdb = self.vis_dyn_bias[0]
		vis_dyn_bias = collections.deque(
			[np.dot(self.vis_dyn_bias[i], f).T for i, f in enumerate(frames)],
			maxlen=self.order)

		hdb = self.hid_dyn_bias[0]
		hid_dyn_bias = collections.deque(
			[np.dot(self.hid_dyn_bias[i], f).T for i, f in enumerate(frames)],
			maxlen=self.order)

		visible = frames[0]
		while True:
			hidden = self.hidden_expectation(visible, sum(hid_dyn_bias))
			yield visible, hidden
			visible = self.visible_expectation(bernoulli(hidden), sum(vis_dyn_bias))
			vis_dyn_bias.append(np.dot(vdb, visible))
			hid_dyn_bias.append(np.dot(hdb, visible))


class TemporalTrainer(Trainer):
	'''
	'''

	def __init__(self, rbm, momentum=0.2, l2=0.1, target_sparsity=None):
		'''
		'''
		super(TemporalTrainer, self).__init__(rbm, momentum, l2, target_sparsity)
		self.grad_dyn_vis = np.zeros(rbm.hid_dyn_bias.shape, float)
		self.grad_dyn_hid = np.zeros(rbm.hid_dyn_bias.shape, float)

	def calculate_gradients(self, frames_batch):
		'''Calculate gradients using contrastive divergence.

		Returns a 5-tuple of gradients: weights, visible bias, hidden bias,
		dynamic visible bias, and dynamic hidden bias.

		frames_batch: An (order, visible units, batch size) array containing a
		  batch of frames of visible data.

		  Frames are assumed to be reversed temporally, across the order
		  dimension, i.e., frames_batch[0] is the current visible frame in each
		  batch element, frames_batch[1] is the previous visible frame,
		  frames_batch[2] is the one before that, etc.
		'''
		order, _, batch_size = frames_batch.shape
		assert order == self.rbm.order

		vis_bias = sum(np.dot(self.rbm.vis_dyn_bias[i], f).T for i, f in enumerate(frames_batch))
		hid_bias = sum(np.dot(self.rbm.hid_dyn_bias[i], f).T for i, f in enumerate(frames_batch))

		v0 = frames_batch[0].T
		h0 = self.rbm.hidden_expectation(v0, hid_bias)
		v1 = self.rbm.visible_expectation(bernoulli(h0), vis_bias)
		h1 = self.rbm.hidden_expectation(v1, hid_bias)

		gw = (np.dot(h0.T, v0) - np.dot(h1.T, v1)) / batch_size
		gv = (v0 - v1).mean(axis=0)
		gh = (h0 - h1).mean(axis=0)

		gvd = np.zeros(self.rbm.vis_dyn_bias.shape, float)
		ghd = np.zeros(self.rbm.hid_dyn_bias.shape, float)
		v = v0 - self.rbm.vis_bias - vis_bias
		for i, f in enumerate(frames_batch):
			gvd[i] += np.dot(v.T, f)
			ghd[i] += np.dot(h0.T, f)
		v = v1 - self.rbm.vis_bias - vis_bias
		for i, f in enumerate(frames_batch):
			gvd[i] -= np.dot(v.T, f)
			ghd[i] -= np.dot(h1.T, f)

		return gw, gv, gh, gvd, ghd

	def apply_gradients(self, weights, visible, hidden, visible_dyn, hidden_dyn,
						learning_rate=0.2):
		'''
		'''
		def update(name, g, _g, l2=0):
			target = getattr(self.rbm, name)
			g *= 1 - self.momentum
			g += self.momentum * (g - l2 * target)
			target += learning_rate * g
			_g[:] = g

		update('vis_bias', visible, self.grad_vis)
		update('hid_bias', hidden, self.grad_hid)
		update('weights', weights, self.grad_weights, self.l2)
		update('vis_dyn_bias', visible_dyn, self.grad_vis_dyn, self.l2)
		update('hid_dyn_bias', hidden_dyn, self.grad_hid_dyn, self.l2)


import scipy.signal
convolve = scipy.signal.convolve


class Convolutional(RBM):
	'''
	'''

	def __init__(self, num_filters, filter_shape, pool_shape, binary=True, scale=0.001):
		'''Initialize a convolutional restricted boltzmann machine.

		num_filters: The number of convolution filters.
		filter_shape: An ordered pair describing the shape of the filters.
		pool_shape: An ordered pair describing the shape of the pooling groups.
		binary: True if the visible units are binary, False if the visible units
		  are normally distributed.
		scale: Scale initial values by this parameter.
		'''
		self.num_filters = num_filters

		self.weights = scale * rng.randn(num_filters, *filter_shape)
		self.vis_bias = scale * rng.randn()
		self.hid_bias = scale * rng.randn(num_filters)

		self._visible = binary and sigmoid or identity
		self._pool_shape = pool_shape

	def _pool(self, hidden):
		'''Given activity in the hidden units, pool it into groups.'''
		_, r, c = hidden.shape
		rows, cols = self._pool_shape
		active = np.exp(hidden.T)
		pool = np.zeros(active.shape, float)
		for j in range(int(np.ceil(float(c) / cols))):
			cslice = slice(j * cols, (j + 1) * cols)
			for i in range(int(np.ceil(float(r) / rows))):
				mask = (cslice, slice(i * rows, (i + 1) * rows))
				pool[mask] = active[mask].sum(axis=0).sum(axis=0)
		return pool.T

	def pooled_expectation(self, visible, bias=0.):
		'''Given visible data, return the expected pooling unit values.'''
		activation = np.exp(np.array([
			convolve(visible, self.weights[k, ::-1, ::-1], 'valid')
			for k in range(self.num_filters)]).T + self.hid_bias + bias).T
		return 1. - 1. / (1. + self._pool(activation))

	def hidden_expectation(self, visible, bias=0.):
		'''Given visible data, return the expected hidden unit values.'''
		activation = np.exp(np.array([
			convolve(visible, self.weights[k, ::-1, ::-1], 'valid')
			for k in range(self.num_filters)]).T + self.hid_bias + bias).T
		return activation / (1. + self._pool(activation))

	def visible_expectation(self, hidden, bias=0.):
		'''Given hidden states, return the expected visible unit values.'''
		activation = sum(
			convolve(hidden[k], self.weights[k], 'full')
			for k in range(self.num_filters)) + self.vis_bias + bias
		return self._visible(activation)


class ConvolutionalTrainer(Trainer):
	'''
	'''

	def calculate_gradients(self, visible):
		'''Calculate gradients for an instance of visible data.

		Returns a 3-tuple of gradients: weights, visible bias, hidden bias.

		visible: A single array of visible data.
		'''
		v0 = visible
		h0 = self.rbm.hidden_expectation(v0)
		v1 = self.rbm.visible_expectation(bernoulli(h0))
		h1 = self.rbm.hidden_expectation(v1)

		# h0.shape == h1.shape == (num_filters, visible_rows - filter_rows + 1, visible_columns - filter_columns + 1)
		# v0.shape == v1.shape == (visible_rows, visible_columns)

		gw = np.array([
			convolve(v0, h0[k, ::-1, ::-1], 'valid') -
			convolve(v1, h1[k, ::-1, ::-1], 'valid')
			for k in range(self.rbm.num_filters)])
		gv = (v0 - v1).sum()
		gh = (h0 - h1).sum(axis=-1).sum(axis=-1)
		if self.target_sparsity is not None:
			h = self.target_sparsity - self.rbm.hidden_expectation(visible).mean(axis=-1).mean(axis=-1)

		logging.debug('displacement: %.3g, hidden activations: %.3g',
					  np.linalg.norm(gv), h0.mean(axis=-1).mean(axis=-1).std())

		return gw, gv, gh


class MeanCovariance(RBM):
	'''
	'''

	def __init__(self, num_visible, num_mean, num_precision, scale=0.001):
		'''Initialize a mean-covariance restricted boltzmann machine.

		num_visible: The number of visible units.
		num_mean: The number of units in the hidden mean vector.
		num_precision: The number of units in the hidden precision vector.
		'''
		super(MeanCovariance, self).__init__(
			num_visible, num_mean, binary=False, scale=scale)

		# replace the hidden bias to reflect the precision units.
		self.hid_bias = scale * rng.randn(num_precision, 1)

		self.hid_mean = scale * rng.randn(num_mean, 1)

		self.hid_factor_u = scale * -abs(rng.randn(num_precision - 1))
		self.hid_factor_c = scale * -abs(rng.randn(num_precision))
		self.hid_factor_l = scale * -abs(rng.randn(num_precision - 1))

		self.vis_factor = scale * rng.randn(num_visible, num_precision)

	@property
	def hid_factor(self):
		return (numpy.diag(self.hid_factor_u, 1) +
				numpy.diag(self.hid_factor_c, 0) +
				numpy.diag(self.hid_factor_l, -1))

	def hidden_expectation(self, visible):
		'''Given visible data, return the expected hidden unit values.'''
		z = numpy.dot(visible.T, self.vis_factor)
		return sigmoid(numpy.dot(z * z, self.hid_factor).T + self.hid_bias)

	def visible_expectation(self, hidden):
		'''Given hidden states, return the expected visible unit values.'''
		z = numpy.diag(numpy.dot(-self.hid_factor.T, hidden))
		Sinv = numpy.dot(self.vis_factor, numpy.dot(z, self.vis_factor.T))
		return numpy.dot(numpy.dot(numpy.pinv(Sinv), self.weights), self.hid_mean)


class MeanCovarianceTrainer(Trainer):
	'''
	'''



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import datetime # Supplies classes for manipulating dates and times in both simple and complex ways
import os.path # The path module suitable for the operating system Python is running on, and therefore usable for local paths
import pysrt # SubRip (.srt) subtitle parser and writer
import multiprocessing # A package that supports spawning processes using an API similar to the threading module.
import time # Provides various time-related functions.
from cerebrum.language.utilities import LanguageMemoryUtil # BUILT-IN Memory operations package

class LanguageAnalyzer():

	@staticmethod
	def word_to_phones(word):
		with open("cerebrum/language/dictionaries/en/cmudict-0.7b.txt") as infile:
			for row in infile:
				if row.split()[0] == word.upper():
					return " ".join(row.split()[1:])
			return ""

	#MAIN CODE BLOCK
	@staticmethod
	def start(text_input,language_analysis_stimulated):
		time.sleep(0.3) # Wait 0.5 seconds for other processes's start
		t0 = time.time() # Initiation time
		if os.path.exists(text_input): # If captions file exist
			subs = pysrt.open(text_input) # Get whole subtitles
			i = 0 # Step counter
			while i < len(subs): # While step counter less than amount of subtitles
				time.sleep(0.5) # Wait 0.5 seconds to prevent aggressive loop
				if (time.time() - t0) > subs[i].start.seconds: # If current time is greater than subtitle's start
					sub_starting_time = datetime.datetime.now() # Starting time of the memory
					language_analysis_stimulated.value = 1 # Language analysis stimulated
					sub_ending_time = sub_starting_time + datetime.timedelta(seconds=(subs[i].end - subs[i].start).seconds) # Calculate the ending time by subtitle's delta
					sub = subs[i].text.encode('ascii','ignore') # Encode subtitle's text by ascii and assign to sub variable
					sub = sub.translate(None, '!@#$?,')
					words = sub.split()
					phone_groups = []
					for word in words:
						phone_groups.append(LanguageAnalyzer.word_to_phones(word))
					phones = " ".join(phone_groups)
					phone_duration = datetime.timedelta(seconds=(subs[i].end - subs[i].start).seconds) / len(phones)
					starting_time = sub_starting_time
					for word_inphones in phone_groups:
						ending_time = starting_time + phone_duration * len(word_inphones.split())
						if ending_time <= sub_ending_time and word_inphones != "":
							process5 = multiprocessing.Process(target=LanguageMemoryUtil.add_memory, args=(word_inphones, starting_time, ending_time)) # Define write memory process
							process5.start() # Start write memory process
						starting_time = ending_time + datetime.timedelta(milliseconds=50)
					print subs[i].text + "\n" # Print subtitle's text
					print phones + "\n"
					print "-------------------------------------------------------------\n"

					language_analysis_stimulated.value = 0 # Language analysis NOT stimulated
					i  += 1 # Increase step counter
		else: # If captions file doesn't exist
			raise ValueError('VTT file doesn\'t exist!') # Raise a ValueError



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import datetime
import os.path
import os

HIPPOCAMPUS = os.path.expanduser("~/Hippocampus/")

if os.path.exists(HIPPOCAMPUS):
	os.remove(HIPPOCAMPUS)
	print HIPPOCAMPUS + " is removed"



POTENTIAL_RANGE = 110000 # Resting potential: -70 mV Membrane potential range: +40 mV to -70 mV --- Difference: 110 mV = 110000 microVolt --- https://en.wikipedia.org/wiki/Membrane_potential
ACTION_POTENTIAL = 15000 # Resting potential: -70 mV Action potential: -55 mV --- Difference: 15mV = 15000 microVolt --- https://faculty.washington.edu/chudler/ap.html
AVERAGE_SYNAPSES_PER_NEURON = 8200 # The average number of synapses per neuron: 8,200 --- http://www.ncbi.nlm.nih.gov/pubmed/2778101

# https://en.wikipedia.org/wiki/Neuron

class Neuron():

	total_neurons = 0

	def __init__(self, dendritic_spine_count=AVERAGE_SYNAPSES_PER_NEURON/2, axon_terminal_count=AVERAGE_SYNAPSES_PER_NEURON/2):
		self.dendritic_spine_count = dendritic_spine_count
		self.axon_terminal_count = axon_terminal_count
		self.neuron_potential = 0
		self.dendrit_constructor(self.dendritic_spine_count)
		self.axon_constructor(self.axon_terminal_count)

		Neuron.total_neurons += 1


	def dendrit_constructor(self, dendritic_spine_count):
		self.dendritic_spines = []
		for i in xrange(dendritic_spine_count):
			self.dendritic_spines.append(DendriticSpine())
		print len(self.dendritic_spines)

	def axon_constructor(self, axon_terminal_count):
		self.axon_terminals = []
		for i in xrange(axon_terminal_count):
			self.axon_terminals.append(AxonTerminal())
		print len(self.axon_terminals)

	def activate(self):
		while True:
			for dendritic_spine in self.dendritic_spines:
				dendritic_spine.potential = dendritic_spine.axon_terminal.potential
				print dendritic_spine.potential
				self.neuron_potential += dendritic_spine.potential
			terminal_potential = self.neuron_potential / len(axon_terminals)
			for axon_terminal in self.axon_terminals:
				axon_terminal.potential = terminal_potential

# https://en.wikipedia.org/wiki/Dendritic_spine

class DendriticSpine():

	def __init__(self, axon_terminal=None, excitement=0, potential=0):
		if isinstance(axon_terminal, DendriticSpine):
			self.axon_terminal = None
		else:
			self.axon_terminal = axon_terminal
		self.potential = potential
		self.excitement = excitement

# https://en.wikipedia.org/wiki/Dendritic_spine

class AxonTerminal():

	def __init__(self, dendritic_spine=None, excitement=0, potential=0):
		if isinstance(dendritic_spine, AxonTerminal):
			self.dendritic_spine = None
		else:
			self.dendritic_spine = dendritic_spine
		self.potential = potential
		self.excitement = excitement

# https://faculty.washington.edu/chudler/cells.html

class SensoryNeuron(Neuron):

	def dendrit_constructor(self, dendritic_spine_count):
		self.dendritic_spines = []
		for i in xrange(dendritic_spine_count):
			self.dendritic_spines.append(DendriticSpine())
		self.dendritic_spines = tuple(self.dendritic_spines)
		print len(self.dendritic_spines)



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

from cerebrum.hearing import HearingMemoryUtil # BUILT-IN Hearing Memory operations package
from cerebrum.vision import VisionMemoryUtil # BUILT-IN Vision Memory operations package
from cerebrum.language import LanguageMemoryUtil
import itertools # Implements a number of iterator building blocks inspired by constructs from APL, Haskell, and SML. Each has been recast in a form suitable for Python
from cerebrum.crossmodal.utilities import MapperUtil # BUILT-IN Crossmodal Memory operations package
import time # Provides various time-related functions.

class MapperStarters():

	# A function for checking two time intervals are overlapping or not --- http://stackoverflow.com/questions/35644301/checking-two-time-intervals-are-overlapping-or-not
	@staticmethod
	def overlap(first_inter,second_inter):
		for f,s in ((first_inter,second_inter), (second_inter,first_inter)):
			#Will check both ways
			for time in (f["starting_time"], f["ending_time"]):
				if s["starting_time"] < time < s["ending_time"]:
					return True
			return False

	@staticmethod
	def startHV():
		# Loop over the timestamps coming from HEARING & VISION
		while True:
			time.sleep(5) # Wait 5 seconds to prevent aggressive loop
			hearing_timestamps = HearingMemoryUtil.get_timestamps() # Get hearing timestamps starting from 0th line
			if not hearing_timestamps:
				continue
			vision_timestamps = VisionMemoryUtil.get_timestamps() # Get vision timestamps starting from 0th line
			if not vision_timestamps:
				continue
			pairs =  MapperUtil.get_allpairs() # Get all pairs
			pairs = list(pairs)
			last_pair = None
			if pairs: # Check if Pairs file exists or not
				for pair in reversed(pairs): # Reverse pairs
					if pair['direction'] == "H2V" or pair['direction'] == "V2H": # Is pair related with HV?
						last_pair = pair # Get latest pair from file
			if last_pair is None: # If last pair is None
				last_pair = {} # Create a fake one
				last_pair['timestamp1'] = 0 # Assign lowest values to timestamps
				last_pair['timestamp2'] = 0 # Assign lowest values to timestamps

			# Cheking all possible combinations for overlapping --- http://stackoverflow.com/questions/35644301/checking-two-time-intervals-are-overlapping-or-not
			for (i1,interval1),(i2,interval2) in itertools.product(enumerate(hearing_timestamps),enumerate(vision_timestamps)): # Cartesian product of enumareted input iterables
				if interval1['starting_time'] < last_pair['timestamp1'] or interval1['starting_time'] < last_pair['timestamp2']: # If current hearing timestamp is earlier than last pair
					continue # Then continue
				if interval2['starting_time'] < last_pair['timestamp1'] or interval2['starting_time'] < last_pair['timestamp2']: # If current vision timestamp is earlier than last pair
					continue # Then continoue
				if MapperStarters.overlap(interval1,interval2): # If interval1 and interval2 is overlapping
					MapperUtil.add_pair(interval1['starting_time'], interval2['starting_time'], "H2V") # Write a H2V pair
					MapperUtil.add_pair(interval2['starting_time'], interval1['starting_time'], "V2H") # Write a V2H pair

	@staticmethod
	def startHL():
		# Loop over the timestamps coming from HEARING & LANGUAGE
		while True:
			time.sleep(5) # Wait 5 seconds to prevent aggressive loop
			hearing_timestamps = HearingMemoryUtil.get_timestamps() # Get hearing timestamps starting from 0th line
			if not hearing_timestamps:
				continue
			language_timestamps = LanguageMemoryUtil.get_timestamps() # Get language timestamps starting from 0th line
			if not language_timestamps:
				continue
			pairs =  MapperUtil.get_allpairs() # Get all pairs
			last_pair = None
			pairs = list(pairs)
			if pairs: # Check if Pairs file exists or not
				for pair in reversed(pairs): # Reverse pairs
					if pair['direction'] == "H2L" or pair['direction'] == "L2H": # Is pair related with HL?
						last_pair = pair # Get latest pair from file
			if last_pair is None: # If last pair is None
				last_pair = {} # Create a fake one
				last_pair['timestamp1'] = 0 # Assign lowest values to timestamps
				last_pair['timestamp2'] = 0 # Assign lowest values to timestamps

			# Cheking all possible combinations for overlapping --- http://stackoverflow.com/questions/35644301/checking-two-time-intervals-are-overlapping-or-not
			for (i1,interval1),(i2,interval2) in itertools.product(enumerate(hearing_timestamps),enumerate(language_timestamps)): # Cartesian product of enumareted input iterables
				if interval1['starting_time'] < last_pair['timestamp1'] or interval1['starting_time'] < last_pair['timestamp2']: # If current hearing timestamp is earlier than last pair
					continue # Then continue
				if interval2['starting_time'] < last_pair['timestamp1'] or interval2['starting_time'] < last_pair['timestamp2']: # If current language timestamp is earlier than last pair
					continue # Then continoue
				if MapperStarters.overlap(interval1,interval2): # If interval1 and interval2 is overlapping
					MapperUtil.add_pair(interval1['starting_time'], interval2['starting_time'], "H2L") # Write a H2L pair
					MapperUtil.add_pair(interval2['starting_time'], interval1['starting_time'], "L2H") # Write a L2H pair

	@staticmethod
	def startVL():
		# Loop over the timestamps coming from VISION & LANGUAGE
		while True:
			time.sleep(5) # Wait 5 seconds to prevent aggressive loop
			vision_timestamps = VisionMemoryUtil.get_timestamps() # Get vision timestamps starting from 0th line
			if not vision_timestamps:
				continue
			language_timestamps = LanguageMemoryUtil.get_timestamps() # Get language timestamps starting from 0th line
			if not language_timestamps:
				continue
			pairs =  MapperUtil.get_allpairs() # Get all pairs
			last_pair = None
			pairs = list(pairs)
			if pairs: # Check if Pairs file exists or not
				for pair in reversed(pairs): # Reverse pairs
					if pair['direction'] == "V2L" or pair['direction'] == "L2V": # Is pair related with VL?
						last_pair = pair # Get latest pair from file
			if last_pair is None: # If last pair is None
				last_pair = {} # Create a fake one
				last_pair['timestamp1'] = 0 # Assign lowest values to timestamps
				last_pair['timestamp2'] = 0 # Assign lowest values to timestamps

			# Cheking all possible combinations for overlapping --- http://stackoverflow.com/questions/35644301/checking-two-time-intervals-are-overlapping-or-not
			for (i1,interval1),(i2,interval2) in itertools.product(enumerate(vision_timestamps),enumerate(language_timestamps)): # Cartesian product of enumareted input iterables
				if interval1['starting_time'] < last_pair['timestamp1'] or interval1['starting_time'] < last_pair['timestamp2']: # If current vision timestamp is earlier than last pair
					continue # Then continue
				if interval2['starting_time'] < last_pair['timestamp1'] or interval2['starting_time'] < last_pair['timestamp2']: # If current language timestamp is earlier than last pair
					continue # Then continoue
				if MapperStarters.overlap(interval1,interval2): # If interval1 and interval2 is overlapping
					MapperUtil.add_pair(interval1['starting_time'], interval2['starting_time'], "V2L") # Write a V2L pair
					MapperUtil.add_pair(interval2['starting_time'], interval1['starting_time'], "L2V") # Write a L2V pair



__author__ = 'Mehmet Mert Yildiran, mert.yildiran@bil.omu.edu.tr'

import datetime # Supplies classes for manipulating dates and times in both simple and complex ways.
import imutils # A series of convenience functions to make basic image processing functions such as translation, rotation, resizing, skeletonization etc.
import time # Provides various time-related functions.
import cv2 # (Open Source Computer Vision) is a library of programming functions mainly aimed at real-time computer vision.
import numpy # The fundamental package for scientific computing with Python.
import os # Provides a portable way of using operating system dependent functionality.
import multiprocessing # A package that supports spawning processes using an API similar to the threading module.
from cerebrum.vision.utilities import VisionMemoryUtil # BUILT-IN Memory operations package
import random # Pseudo-random number generators for various distributions.
import Tkinter

STABILIZATION_DETECTION = 5 # Number of frames to detect stabilization
NON_STATIONARY_PERCENTAGE = 70 # Percentage of frame for detecting NON-STATIONARY CAMERA. Like: ( height * width * float(X) / float(100) )
NON_ZERO_PERCENTAGE = 0 #  Percentage of frame(threshold) for detecting unnecessary movement
TARGET_HEIGHT = 360 # Number of horizontal lines for target video and processing. Like 720p, 360p etc.
MIN_AREA = 500 # Minimum area in square pixels to detect a motion
root = Tkinter.Tk()
SCREEN_WIDTH = root.winfo_screenwidth()
SCREEN_HEIGHT = root.winfo_screenheight()

class VisionPerception():

	# MAIN CODE BLOCK
	@staticmethod
	def start(video_input, vision_perception_stimulated):

		if video_input == "0": # If the video_input is None, then we are reading from webcam
			camera = cv2.VideoCapture(0)
			time.sleep(0.25)
		else: # Otherwise, we are reading from a video file
			time.sleep(0.25)
			camera = cv2.VideoCapture(video_input)

		referenceFrame = None # Initialize the reference frame in the video stream

		starting_time = None
		memory_data_thresh = []
		memory_data_frameDeltaColored = []

		(grabbed, first_frame) = camera.read() # Grab the first frame

		height, width = first_frame.shape[:2] # Get video height and width  from first frame(size)
		#if not height == 720 or not width == 1280:
		if float(width) / float(height) != float(16) / float(9):
			if video_input == "0":
				# There is a STUPIDTY in here
				pass
			else:
				raise ValueError('Aspect ratio of input stream must be [16:9]')
			#warnings.warn("Aspect ratio of input stream must be [16:9]")

		frame_counter = 1 # Define frame counter variable
		motion_detected = 0 # Delta situation checking variable
		delta_value_stack = [] # List of delta values
		non_stationary_camera = 0
		motion_counter = 0
		nonzero_toolow = 0

		beginning_of_stream = datetime.datetime.now()
		while True: # Loop over the frames of the video

			(grabbed, frame) = camera.read() # Grab the current frame and initialize the occupied/unoccupied
			if not grabbed: # If the frame could not be grabbed, then we have reached the end of the video
				break
			frame_counter += 1 # Increase frame counter's value

			if video_input == "0":
				# If we are capturing from camera fuck Time Correction, there is also a STUPIDTY in here
				pass
			else:
				# -------------------- TIME CORRECTION --------------------
				time_delta = datetime.datetime.now() - beginning_of_stream
				current_time_of_realworld = time_delta.seconds + time_delta.microseconds / float(1000000)
				current_time_of_stream = frame_counter / camera.get(cv2.cv.CV_CAP_PROP_FPS)
				diff_of_time = current_time_of_stream - current_time_of_realworld
				if abs(diff_of_time) > (1 / camera.get(cv2.cv.CV_CAP_PROP_FPS)):
					if diff_of_time > 0:
						time.sleep(1 / camera.get(cv2.cv.CV_CAP_PROP_FPS))
					else:
						(grabbed, frame) = camera.read() # Grab the current frame and initialize the occupied/unoccupied
						if not grabbed: # If the frame could not be grabbed, then we have reached the end of the video
							break
						frame_counter += 1 # Increase frame counter's value
						continue
				# -------------------- TIME CORRECTION --------------------

			delta_value = 0 # Delta Value for storing max continuous contour area for current frame

			frame = imutils.resize(frame, height=TARGET_HEIGHT) # Resize frame to 360p. Alternative resizing method:
			height, width = frame.shape[:2] # Get video height and width  from first frame(size)

			gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY) # Convert frame to grayscale

			gray = cv2.bilateralFilter(gray,9,75,75) # Blur current frame with Bilateral Filter for noise reduction

			if referenceFrame is None: # If Reference Frame is None, initialize it
				referenceFrame = gray
				continue

			frameDelta = cv2.absdiff(referenceFrame, gray) # Compute the absolute difference between the current frame and reference frame
			thresh = cv2.threshold(frameDelta, 12, 255, cv2.THRESH_BINARY)[1] # Apply OpenCV's threshold function to get binary frame

			thresh = cv2.dilate(thresh, None, iterations=1) # Dilate the thresholded image to fill in holes
			frameDeltaColored = cv2.bitwise_and(frame,frame, mask= thresh) # Bitwise and - to get delta frame

			# Find contours on thresholded image
			(cnts, _) = cv2.findContours(thresh.copy(), cv2.RETR_EXTERNAL,
				cv2.CHAIN_APPROX_SIMPLE)

			contour_area_stack = [] # List of contour areas's values

			# Loop over the contours
			if cnts:
				for c in cnts: # Contour in Contours
					contour_area_stack.append(cv2.contourArea(c)) # Calculate contour area and append to contour stack
					if cv2.contourArea(c) > MIN_AREA: # If contour area greater than min area
						(x, y, w, h) = cv2.boundingRect(c) # Compute the bounding box for this contour
						cv2.rectangle(frame, (x, y), (x + w, y + h), (0, 255, 0), 2) # Draw it on the frame
				delta_value = max(contour_area_stack) # Assign max contour area to delta value

				if delta_value > MIN_AREA: # If max contour area (delta value) greater than min area
					motion_detected = 1 # Initialize delta situation

				if delta_value > (height * width * float(NON_STATIONARY_PERCENTAGE) / float(100)): # If delta value is too much
					non_stationary_camera = 1
					status_text = "WARNING: NON-STATIONARY CAMERA"
					frameDeltaColored = numpy.zeros_like(frame)
				else:
					non_stationary_camera = 0

				if cv2.countNonZero(thresh) < (height * width * float(NON_ZERO_PERCENTAGE) / float(100)): # If Non Zero count is too low
					nonzero_toolow = 1
					status_text = "WARNING: NON-ZERO TOO LOW"
					frameDeltaColored = numpy.zeros_like(frame)
				else:
					nonzero_toolow = 0

			if motion_detected: # If we are on delta situation

				if starting_time is None:
					starting_time = datetime.datetime.now() # Starting time of the memory
					vision_perception_stimulated.value = 1 # Vision perception stimulated

				if random.randint(0,2) == 1: # IMPORTANT
					memory_data_thresh.append(thresh.tostring())
					memory_data_frameDeltaColored.append(frameDeltaColored.tostring())
					#print type(memory_data_thresh[0])

				if not non_stationary_camera:
					status_text = "MOTION DETECTED"
				delta_value_stack.append(delta_value) # Append max contour area (delta value) to delta value stack

				if len(delta_value_stack) >= STABILIZATION_DETECTION: # If length of delta value stack is greater than or equal to STABILIZATION_DETECTION constant
					delta_value_stack.pop(0) # Pop first element of delta value stack
					# If minimum delta value is greater than (mean of last 5 frame - minimum area / 2) and maximum delta value is less than (mean of last 5 frame + minimum area / 2)
					if min(delta_value_stack) > (numpy.mean(delta_value_stack) - MIN_AREA / 2) and max(delta_value_stack) < (numpy.mean(delta_value_stack) + MIN_AREA / 2):
						ending_time = datetime.datetime.now() # Ending time of the memory
						vision_perception_stimulated.value = 0 # Vision perception NOT stimulated

						if memory_data_thresh and memory_data_frameDeltaColored:
							process4 = multiprocessing.Process(target=VisionMemoryUtil.add_memory, args=(memory_data_thresh, memory_data_frameDeltaColored, starting_time, ending_time)) # Define write memory process
							process4.start() # Start write memory process
						memory_data_thresh = []
						memory_data_frameDeltaColored = []
						starting_time = None

						motion_detected = 0 # Then video STABILIZED
						delta_value_stack = [] # Empty delta value stack
						referenceFrame = None  # Clear reference frame
						if not non_stationary_camera and not nonzero_toolow:
							motion_counter += 1
			else:
				if not non_stationary_camera and not nonzero_toolow:
					status_text = "MOTION UNDETECTED"
					frameDeltaColored = numpy.zeros_like(frame)

			# Draw the text and timestamp on the frame
			cv2.putText(frame, "Diff    : {}".format(delta_value), (10, 20), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 1)
			cv2.putText(frame, "Thresh : {}".format(MIN_AREA), (10, 40), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 1)
			cv2.putText(frame, "Frame : {}".format(frame_counter), (10, 60), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 1)
			cv2.putText(frame, "Status  : {}".format(status_text), (10, 80), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 2)
			cv2.putText(frame, datetime.datetime.now().strftime("%A %d %B %Y %I:%M:%S%p"), (10, frame.shape[0] - 10), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 1)

			# Show the frames and record if the user presses ESC or q
			cv2.imshow("Original Frame", frame)
			cv2.moveWindow("Original Frame",50 * SCREEN_WIDTH / 1920,100 * SCREEN_HEIGHT / 1080)
			cv2.imshow("Frame Threshhold", thresh)
			cv2.moveWindow("Frame Threshhold",50 * SCREEN_WIDTH / 1920,550 * SCREEN_HEIGHT / 1080)
			cv2.imshow("Frame Delta", frameDelta)
			cv2.moveWindow("Frame Delta",1200 * SCREEN_WIDTH / 1920,550 * SCREEN_HEIGHT / 1080)
			cv2.imshow("Frame Delta Colored", frameDeltaColored)
			cv2.moveWindow("Frame Delta Colored",1200 * SCREEN_WIDTH / 1920,100 * SCREEN_HEIGHT / 1080)
			key = cv2.waitKey(1) & 0xFF

			# if the `ESC` or `q` key is pressed, break the loop
			if key == ord("q") or key == ord("\x1b"):
				os.system("killall python") # Temporary line for practicality in DEVELOPMENT
				break

		cv2.destroyAllWindows() # Close any open windows
		camera.release() # Release the capture device
